// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#import  <objc/Object.h>
#include <X11/Xlib.h>
#include <IMdkit.h>
#include <Xi18n.h>
#import  "Intrinsic.h"
#import  "KText.h"

#include "types.h"

@class KAutomataForX;
@class KeyTranslator;

typedef struct {
	XRectangle area;			// rectangle to display the widget in.
	XRectangle area_needed;		// constraint imposed by the client.
	XPoint spot_location;
	Colormap colormap;			// default required.
	Atom std_colormap;
	unsigned long foreground;
	unsigned long background;
	Pixmap background_pixmap;
	XOC font_set;				// output context.  always specified.
	XRectangle max_logical_extent;
	int line_space;
	Cursor cursor;

	BOOL foreground_set;		// YES if foreground has been set.
	BOOL background_set;		// YES if background has bedn set.

	Widget widget;
} SubAttr;

@interface IContext : Object
{
	ICIDType icid;				// IC id
	ConnectIDType connect_id;
	
	// IC attributes
	XIMStyle input_style;		// once set, not changed.
	Window client_window;		// once set, not changed.
	Window focus_window;		// changed freely.  default: client_window
	char *resource_name;		// changed.
	char *resource_class;		// changed.
	EventMask filter_events;	// never set.
	
	SubAttr preedit_attr;		// preedit sub-attributes
	SubAttr status_attr;		// status sub-attributes

	BOOL line_mode;
	KeyboardType keyboard;
	char *kdf_name;
	char *ksm_name;
	
	Widget ebox;				// box to contain text and status.
	Widget editShell;
	id text;
	Widget statusShell;
	Widget status;
	
	KAutomataForX *am;			// used in character mode.
	KeyTranslator *kt;			//          ''
}

- init;							// -init must not be called.
- initWith:(IMChangeICStruct *)call_data;
- free;
- setID:(ICIDType)newIcid;
- (ICIDType)getID;
- changeParameters;
- insertString:(char *)string;

// protocol methods

- (int)setICFocus:(IMChangeFocusStruct *)data;
- (int)unsetICFocus:(IMChangeFocusStruct *)data;
- (int)resetIC:(IMResetICStruct *)data;
- (int)setICValues:(IMChangeICStruct *)data;
- (int)getICValues:(IMChangeICStruct *)data;
- (int)triggerNotify:(IMTriggerNotifyStruct *)data;
- (int)forwardEvent:(IMForwardEventStruct *)data;
- (int)preeditStartReply:(IMPreeditCBStruct *)data;
- (int)preeditCaretReply:(IMPreeditCBStruct *)data;
- (int)strConversionReply:(IMStrConvCBStruct *)data;

//
//  private methods
//

- charModeConversion;
- convertComposingCharTo:(char *)chinese;
- startPreedit;
- endPreedit;
- changeStatusLabel;
- (char *)getComposingChar;
- driveAutomata:(IMForwardEventStruct *)data;
- flushAutomata;
- flushText;
- commitString:(char *)wansung_string;
- commitJohabString:(char *)johab_string;
- setAlreadySetAttributes;
- setAlreadySetSubattributesOf:(SubAttr *)subattr;
- popupControlPanel:(XPointer)data;
- changeTextMaxWidth;
- setCharModeTextWidth;

// IC attribute methods

- setNinputStyle:(XICAttribute *)attr;
- getNinputStyle:(XICAttribute *)attr;
- setNclientWindow:(XICAttribute *)attr;
- getNclientWindow:(XICAttribute *)attr;
- setNfocusWindow:(XICAttribute *)attr;
- getNfocusWindow:(XICAttribute *)attr;
- setNresourceName:(XICAttribute *)attr;
- getNresourceName:(XICAttribute *)attr;
- setNresourceClass:(XICAttribute *)attr;
- getNresourceClass:(XICAttribute *)attr;
- setNfilterEvents:(XICAttribute *)attr; // never called
- getNfilterEvents:(XICAttribute *)attr;

// preedit & status sub-attribute methods

- setNarea:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNarea:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNareaNeeded:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNareaNeeded:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNspotLocation:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNspotLocation:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNcolorMap:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNcolorMap:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNstdColorMap:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNstdColorMap:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNforeground:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNforeground:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNbackground:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNbackground:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNbackgroundPixmap:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNbackgroundPixmap:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNfontSet:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNfontSet:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNlineSpace:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNlineSpace:(XICAttribute *)attr of:(SubAttr *)subattr;
- setNcursor:(XICAttribute *)attr of:(SubAttr *)subattr;
- getNcursor:(XICAttribute *)attr of:(SubAttr *)subattr;

@end
