// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#import  <objc/Object.h>
#define  XK_KOREAN
#include <X11/keysym.h>

#define MAXKEYSPERCHAR 10
#define MAXCANDIDATES  6

typedef enum {
	NSTART,
	NCHO1 , NCHO2 , NCHO3 , NCHO4 ,
	NJUNG1, NJUNG2, NJUNG3, NJUNG4,
	NJONG1, NJONG2, NJONG3, NJONG4
} CompositionState;

typedef unsigned int JamoKey;

typedef struct {
	JamoKey jamo_key;
	int state_offset;
} JamoKeyAndStateOffset;
	
enum { JUSTFINE, CHARBREAK, EMPTY, WRONGKEY, OVERFLOW };

@interface KAutomataCore : Object
{
@private
	char composed_char[3];		// in 2 byte johab code
	char composing_char[3];		// in 2 byte johab code
	CompositionState state;
	int num_elected_keys;
	JamoKey elected_keys[MAXKEYSPERCHAR+1];
	CompositionState history[MAXKEYSPERCHAR+1];
	JamoKeyAndStateOffset cand_list_save[MAXKEYSPERCHAR+1][MAXCANDIDATES];
}

- init;

// keyCandidateList is terminated by an element whose jamo_key member
// is 0 if the number of elements is less than MAXCANDIDATES.
// If the number is MAXCANDIDATES, no terminating sign is needed.
- (int)eatKey:(JamoKeyAndStateOffset *)keyCandidateList;
- (int)deleteLastKey;
- (char *)getComposedChar;
- (char *)getComposingChar;
- clear;

// private methods

- reset;

@end
