// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#include "Intrinsic.h"
#import  <objc/Object.h>
#include <X11/Xaw/AsciiText.h>

@class KAutomataForX;
@class KeyTranslator;
@class XChineseDic;

struct trigger_key {
	KeySym keysym;
	unsigned int modifier;
	unsigned int modifier_mask;
};

typedef enum { MRAW, MTRANS, MKOREAN } KeyboardType;

@interface KText : Object
{
	KeyboardType keyboard;
	KAutomataForX *am;
	BOOL phantom;				// indicates that before the caret is a
								// phantom character.
	KeyTranslator *kt;
	Widget text;
	EventMask text_event_mask;
	Widget translator;			// Widget on which text translations are
								// installed.
	XChineseDic *cd;
	XawTextPosition conv_start_pos;
	XawTextPosition conv_end_pos;
	
	struct trigger_key toggle;
	struct trigger_key conversion;
}

- init;							// this must not be called.
- initFromKDF:(char *)kdf_path KSM:(char *)ksm_path
   withWidget:(Widget)text translator:(Widget)tw;
- free;

- setKDF:(char *)path;
- setKSM:(char *)path;
- setToggleKey:(struct trigger_key *)key;
- setConversionKey:(struct trigger_key *)key;
- setChineseDic:(id)cd;
- setKeyboard:(KeyboardType)keyboard;
- (Widget)getWidget;
- setString:(char *)string;
- (char *)getString;
	
//
//  private methods
//

// Text widget action methods

- (void)textForwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textForwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textForwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBeginningOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNextLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPreviousLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNextPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPreviousPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBeginningOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEndOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textScrollOneLineUp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textScrollOneLineDown:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteNextCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeletePreviousCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteNextWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeletePreviousWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillToEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillToEndOfParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewlineAndIndent:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewlineAndBackup:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewline:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectAll:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectSave:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textRedrawDisplay:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSearch:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertChar:(Widget)w :(XKeyEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertString:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFocusIn:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFocusOut:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEnterWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textLeaveWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDisplayCaret:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textMultiply:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFormParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textTransposeCharacters:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNoOp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertFileAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDoSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDoReplaceAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSetField:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPopdownSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textReconnectIm:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;

- deletePhantomIfPresent;
- insertPhantom;
- insertComposingChar;
- beginConversion;
- tryConversionFrom:(XawTextPosition)from to:(XawTextPosition)to;
- convertTo:(char *)chinese;
	
@end

// This function must not be called.
void _KTextEventHandler(Widget w, KText *self, XEvent *ev, Boolean *cont);
