/* -*- mode: c; tab-width: 4 -*- */

/*
 *  This file is part of Byeoroo, a Korean input method server for X11.
 *  Copyright (C) 1997  Park, Jae-hyon.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License , or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  You may contact the author by:
 *     e-mail:  jhpark@entropy.kaist.ac.kr
 *       mail:  Park, Jae-hyon
 *              Physics Department, 305-701
 *              Korea Advanced Institute of Science and Technology
 *              Taejon, South Korea
 */


#include "hangul.h"
#include "XJohab.h"

#define STRBUFLEN 64

int XTextWidthJohab(XFontStruct *fs, XFontStruct *jfs,
					char *s, int n)
{
	int width = 0;
	XChar2b buf[STRBUFLEN];
	char *p;

	for (; n > 0; s = p) {
		if (*s & CMSB) {		/* Korean */
			int l;
			
			for (p = s; n > 0 && p - s < STRBUFLEN*2/3 && *p & CMSB;
				 p += 2, n -= 2)
				;
			l = convert_johab_to_display(s, buf, p - s);
			width += XTextWidth16(jfs, buf, l);
		}
		else {					/* ASCII */
			for (p = s; n > 0 && !(*p & CMSB); p++, n--)
				;
			width += XTextWidth(fs, s, p - s);
		}
	}

	return width;
}

void XDrawStringJohab(Display *display, Drawable d, GC gc, GC jgc,
					  XFontStruct *fs, XFontStruct *jfs,
					  int x, int y, char *s, int n)
{
	XChar2b buf[STRBUFLEN];
	char *p;

	for (; n > 0; s = p) {
		if (*s & CMSB) {		/* Korean */
			int l;
			
			for (p = s; n > 0 && p - s < STRBUFLEN*2/3 && *p & CMSB;
				 p += 2, n -= 2)
				;
			l = convert_johab_to_display(s, buf, p - s);
			XDrawString16(display, d, jgc, x, y, buf, l);
			x += XTextWidth16(jfs, buf, l);
		}
		else {
			for (p = s; n > 0 && !(*p & CMSB); p++, n--)
				;
			XDrawString(display, d, gc, x, y, s, p - s);
			x += XTextWidth(fs, s, p - s);
		}
	}
}
