/* -*- mode: objc; tab-width: 4 -*- */

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#ifndef _BYEOROO_H_
#define _BYEOROO_H_

#include <X11/Xlib.h>
#include <IMdkit.h>
#include <Xi18n.h>
#include "Intrinsic.h"
#import  "IContext.h"
#import  "XChineseDic.h"
#include "types.h"

#define APPCLASS  "Byeoroo"
#define IMSNAME   "byeoroo"
#define IMSLOCALE "ko_KR"
#define FILTEREVENTMASK KeyPressMask

#define SPECIAL_IN_JOHAB "\xcb\x62"

#define MIN(a, b) ((a) < (b) ? (a) : (b))

struct app_data {
	char *transport;
	char *toggle_key;
	char *chinese_key;
	char *popup_key;
	char *flush_key;
	char *search_path;
	char *default_kdf;
	int number_of_menu_items;
};

enum { TOGGLE, POPUP, CONVERSION };	// trigger key index

extern struct app_data app_data;
extern XtAppContext app_context;
extern XIMS ims;
extern Display *display;
extern XOM om;
extern IContext *focusedIC;
extern IContext *lastFocusedIC;
extern XChineseDic *cd;
extern BOOL cp_popped_up;
extern XIMTriggerKey trigger_keys[];
extern XIMTriggerKey flush_key;

// Main window widgets
extern Widget topLevel;
extern Widget box;	
extern Widget korEngStat;
extern Widget charLineStat;
extern Widget kdfLabel;
extern Widget ksmLabel;

// Control panel widgets
extern Widget cpshell;
extern Widget cpform;
extern Widget cptitle;
extern Widget cplabel1;
extern Widget cplabel2;
extern Widget cplabel3;
extern Widget cplabel4;
extern Widget cpchar;
extern Widget cpline;
extern Widget cpkdf;
extern Widget cpksm;
extern Widget cpcode;
extern Widget cpsample;
extern Widget cpokay;
extern Widget cpminput;
extern Widget cpcancel;
extern Widget cplistshell;
extern Widget cplist;

// Menu widgets
extern Widget menu;
extern Widget strshell;
extern Widget strlist;

void install_actions();
void install_translations();
void app_main_loop(XtAppContext ac, BOOL *continue_looping);
int trigger_key_index(XKeyEvent *kev);
BOOL is_trigger_key(XKeyEvent *kev, XIMTriggerKey *trigger);
void print_e(int exit_code, char *format, ...);
void CenterWidgetAtPoint(Widget w, Position x, Position y);
void SetListWidth(Widget w);

#endif /* _BYEOROO_H_ */
