// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include "Intrinsic.h"
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include "JohabLabel.h"

#import  "ICManager.h"
#import  "IContext.h"
#import  "KAutomataForX.h"
#import  "KeyTranslator.h"
#include "textActions.h"
#import  "XChineseDic.h"

#include "byeoroo.h"
#include "sutil.h"

#define CODELEN 4

#define XtNtransport "transport"
#define XtCTransport "Transport"
#define XtNkorEngToggleKey "korEngToggleKey"
#define XtCKorEngToggleKey "KorEngToggleKey"
#define XtNchineseConversionKey "chineseConversionKey"
#define XtCChineseConversionKey "ChineseConversionKey"
#define XtNpopupKey "popupKey"
#define XtCPopupKey "PopupKey"
#define XtNflushKey "flushKey"
#define XtCFlushKey "FlushKey"
#define XtNsearchPath "searchPath"
#define XtCSearchPath "SearchPath"
#define XtNdefaultKeyboardDescriptionFile "defaultKeyboardDescriptionFile"
#define XtCDefaultKeyboardDescriptionFile "DefaultKeyboardDescriptionFile"
#define XtNnumberOfMenuItems "numberOfMenuItems"
#define XtCNumberOfMenuItems "NumberOfMenuItems"

#define off(f) XtOffsetOf(struct app_data, f)
static XtResource resources[] = {
	{
		XtNtransport,
		XtCTransport,
		XtRString,
		sizeof(char *),
		off(transport),
		XtRImmediate,
		NULL
	},
	{
		XtNkorEngToggleKey,
		XtCKorEngToggleKey,
		XtRString,
		sizeof(char *),
		off(toggle_key),
		XtRImmediate,
		NULL
	},
	{
		XtNchineseConversionKey,
		XtCChineseConversionKey,
		XtRString,
		sizeof(char *),
		off(chinese_key),
		XtRImmediate,
		NULL
	},
	{
		XtNpopupKey,
		XtCPopupKey,
		XtRString,
		sizeof(char *),
		off(popup_key),
		XtRImmediate,
		NULL
	},
	{
		XtNflushKey,
		XtCFlushKey,
		XtRString,
		sizeof(char *),
		off(flush_key),
		XtRImmediate,
		NULL
	},
	{
		XtNsearchPath,
		XtCSearchPath,
		XtRString,
		sizeof(char *),
		off(search_path),
		XtRImmediate,
		NULL
	},
	{
		XtNdefaultKeyboardDescriptionFile,
		XtCDefaultKeyboardDescriptionFile,
		XtRString,
		sizeof(char *),
		off(default_kdf),
		XtRImmediate,
		NULL
	},
	{
		XtNnumberOfMenuItems,
		XtCNumberOfMenuItems,
		XtRInt,
		sizeof(int),
		off(number_of_menu_items),
		XtRImmediate,
		(XtPointer)0
	}
};
#undef off

static XrmOptionDescRec options[] = {
	{ "-transport", "*" XtNtransport, XrmoptionSepArg, NULL },
	{ "-kdf", "*" XtNdefaultKeyboardDescriptionFile, XrmoptionSepArg, NULL },
};

static void WMProtocols(Widget, XEvent *, String *, Cardinal *);
static void ShowSample(Widget, XEvent *, String *, Cardinal *);
		
static XtActionsRec actions[] = {
	{ "WMProtocols", WMProtocols },
	{ "ShowSample", ShowSample }
};

struct app_data app_data;
XtAppContext app_context;
XIMS ims;						// input method service
Window iwin;					// Window for transport
Display *display;				
XOM om;							// output method
IContext *focusedIC = nil;		// currently focused IC
IContext *lastFocusedIC = nil;
XChineseDic *cd;
BOOL cp_popped_up = NO;
XIMTriggerKey trigger_keys[] = {
	{ 0L, 0L, 0L },				// Korean-English key
	{ 0L, 0L, 0L },				// Popup key
	{ 0L, 0L, 0L }				// Chinese conversion key
};
XIMTriggerKey flush_key;		// Line-mode flush key

//
//  main window widgets
//

Widget topLevel;
Widget box;
Widget korEngStat;
Widget charLineStat;
Widget kdfLabel;
Widget ksmLabel;

//
//  control panel widgets
//

Widget cpshell;
Widget cpform;
Widget cptitle;
Widget cplabel1;
Widget cplabel2;
Widget cplabel3;
Widget cplabel4;
Widget cpchar;
Widget cpline;
Widget cpkdf;
Widget cpksm;
Widget cpcode;
Widget cpsample;
Widget cpokay;
Widget cpcinput;
Widget cpminput;
Widget cpcancel;
Widget cplistshell;
Widget cplist;

//
//  menu widgets
//

Widget menu;
Widget strshell;
Widget strlist;

static Atom wm_delete_window;
static Atom wm_protocols;

static XIMTriggerKeys trigger_key_list = {
	XtNumber(trigger_keys),
	trigger_keys
};

static XIMStyle styles[] = {
    XIMPreeditCallbacks | XIMStatusCallbacks,
    XIMPreeditCallbacks | XIMStatusArea,     
    XIMPreeditCallbacks | XIMStatusNothing,  
    XIMPreeditCallbacks | XIMStatusNone,     
    XIMPreeditPosition  | XIMStatusCallbacks,
    XIMPreeditPosition  | XIMStatusArea,     
    XIMPreeditPosition  | XIMStatusNothing,  
    XIMPreeditPosition  | XIMStatusNone,     
    XIMPreeditArea      | XIMStatusCallbacks,
    XIMPreeditArea      | XIMStatusArea,     
    XIMPreeditArea      | XIMStatusNothing,  
    XIMPreeditArea      | XIMStatusNone,     
    XIMPreeditNothing   | XIMStatusCallbacks,
    XIMPreeditNothing   | XIMStatusArea,     
    XIMPreeditNothing   | XIMStatusNothing,  
    XIMPreeditNothing   | XIMStatusNone,     
    XIMPreeditNone      | XIMStatusCallbacks,
    XIMPreeditNone      | XIMStatusArea,     
    XIMPreeditNone      | XIMStatusNothing,  
    XIMPreeditNone      | XIMStatusNone      
};
static XIMStyles style_list = { XtNumber(styles), styles };

static XIMEncoding encodings[] = {
	"COMPOUND_TEXT"				// This one suffices.
};
static XIMEncodings encoding_list = { XtNumber(encodings), encodings };

static char *program_name;		// used when printing an error message.
static ICManager *theICManager;

// These three variables are referenced only in CPPopupFileList(),
// FreeFileList() directly and in CPFileChosen() indirectly.
// (via call_data->string)
static char **file_list = NULL;	// used to construct a file name list widget.
static int file_list_size = 0;
static int nfiles;				// number of file names

// These two variables are referenced in ItemChosen(), FreeStringList(),
// StringChosen().
static char **string_list = NULL;
static int nstrings = 0;

// This is used in create_widgets(),
static char code_str[CODELEN+1] = "";

static void sort_if_necessary(int argc, char *argv[]);
static void create_widgets();
static void open_om();
static void parse_resources();
static void help_msg();

static int protocol_handler(XIMS ims, IMProtocol *call_data);
static int open_handler(XIMS, IMOpenStruct *);
static int close_handler(XIMS, IMCloseStruct *);
static int create_ic_handler(XIMS, IMChangeICStruct *);
static int destroy_ic_handler(XIMS, IMDestroyICStruct *);

static void CPPopupFileList(Widget w, char *cl_d, XtPointer c_d);
static void CPFileChosen(Widget w, XtPointer cl_d, XawListReturnStruct *c_d);
static void CPOkay(Widget w, XtPointer client_data, XtPointer call_data);
static void CPCancel(Widget w, XtPointer client_data, XtPointer call_data);
static void FreeFileList();
static void CPCodeInput(Widget w, XtPointer client_data, XtPointer call_data);
static void ItemChosen(Widget w, Widget client_data, XtPointer call_data);
static void StringChosen(Widget w, XtPointer cl_d, XawListReturnStruct *c_d);
static void FreeStringList();

int main(int argc, char *argv[])
{
	program_name = argv[0];

	sort_if_necessary(argc, argv);
	
	if ((theICManager = [[ICManager alloc] init]) == nil)
		print_e(1, "Cannot manage input contexts.\n");
	
	XtSetLanguageProc(NULL, NULL, NULL);
	topLevel = XtVaOpenApplication(&app_context, APPCLASS,
								   options, XtNumber(options),
								   &argc, argv,
								   NULL, // fallback resources
								   sessionShellWidgetClass,
								   NULL);
	if (argc > 1) {
		print_e(0, "Bad command line option: %s\n", argv[1]);
		help_msg();
		exit(1);
	}
	XtVaGetApplicationResources(topLevel,
								&app_data,
								resources, XtNumber(resources),
								NULL);
	parse_resources();

	cd = [[XChineseDic alloc] initFromSearchPath:app_data.search_path
							  parentWidget:topLevel];
	create_widgets();
	install_text_actions(app_context);
	XtAppAddActions(app_context, actions, XtNumber(actions));
	XtRealizeWidget(topLevel);
	XtRealizeWidget(cpshell);	// to set WM_PROTOCOLS property.
	XtRealizeWidget(strshell);	//              ''
	XtRealizeWidget([cd getShell]);	//          ''
	
	display = XtDisplay(topLevel);
	iwin = XtWindow(topLevel);
	open_om();
	if (!retrieve_keyboard_mapping(display))
		exit(1);
	
	if ((ims = IMOpenIM(display,
						IMModifiers, "Xi18n", // We use R6 XIM protocol.
						IMServerWindow, iwin,
						IMServerName, IMSNAME,
						IMLocale, IMSLOCALE,
						IMServerTransport, app_data.transport,
						IMInputStyles, &style_list,
						IMEncodingList, &encoding_list,
						IMFilterEventMask, FILTEREVENTMASK,
						IMOnKeysList, &trigger_key_list,
						IMProtocolHandler, protocol_handler,
						NULL)) == NULL)
		print_e(1, "Cannot open input method service.\n");
	
    wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
    wm_protocols = XInternAtom(display, "WM_PROTOCOLS", False);
    XSetWMProtocols(display, XtWindow(topLevel), &wm_delete_window,1);
    XSetWMProtocols(display, XtWindow(cpshell), &wm_delete_window,1);
    XSetWMProtocols(display, XtWindow([cd getShell]), &wm_delete_window,1);
	XSetWMProtocols(display, XtWindow(strshell), &wm_delete_window, 1);
	
	XtAppMainLoop(app_context);

	return -5;			  // if control reaches here, something must be wrong.
}

static void sort_if_necessary(int argc, char *argv[])
{
	while (*argv) {
		if (strcmp(*argv, "-sortcd") == 0)
			break;
		argv++;
	}
	if (*argv == NULL)			// sort option not found.
		return;

	argv++;
	if (*argv == NULL) {
		help_msg();
		exit(1);
	}
	if (!has_extension(*argv, CDICEXT)) {
		print_e(0, "%s doesn't have " CDICEXT " extension.\n", *argv);
		exit(1);
	}
	
	exit(!sort_and_index(*argv));
}

static void create_widgets()
{
	XFontStruct *jf;
	Dimension iw, ih;
	int i;
	
	box = XtVaCreateManagedWidget("box", boxWidgetClass, topLevel, NULL);
	korEngStat = XtVaCreateManagedWidget("korEngStat", johabLabelWidgetClass,
										 box, NULL);
	charLineStat = XtVaCreateManagedWidget("charLineStat", labelWidgetClass,
										   box, NULL);
	kdfLabel = XtVaCreateManagedWidget("kdfLabel", labelWidgetClass, box,
									   NULL);
	ksmLabel = XtVaCreateManagedWidget("ksmLabel", labelWidgetClass, box,
									   NULL);

	cpshell = XtVaCreatePopupShell("cpshell", transientShellWidgetClass,
								   topLevel, NULL);
	cpform = XtVaCreateManagedWidget("cpform", formWidgetClass, cpshell, NULL);
	cptitle = XtVaCreateManagedWidget("cptitle", labelWidgetClass, cpform,
									  NULL);
	cplabel1 = XtVaCreateManagedWidget("cplabel1", labelWidgetClass, cpform,
									   NULL);
	cpchar = XtVaCreateManagedWidget("cpchar", toggleWidgetClass, cpform,
									 NULL);
	cpline = XtVaCreateManagedWidget("cpline", toggleWidgetClass, cpform,
									 NULL);
	cplabel2 = XtVaCreateManagedWidget("cplabel2", labelWidgetClass, cpform,
									   NULL);
	cpkdf = XtVaCreateManagedWidget("cpkdf", commandWidgetClass, cpform, NULL);
	cplabel3 = XtVaCreateManagedWidget("cplabel3", labelWidgetClass, cpform,
									   NULL);
	cpksm = XtVaCreateManagedWidget("cpksm", commandWidgetClass, cpform, NULL);

	cplabel4 = XtVaCreateManagedWidget("cplabel4", labelWidgetClass, cpform,
									   NULL);
	cpcode = XtVaCreateManagedWidget("cpcode", asciiTextWidgetClass, cpform,
									 XtNuseStringInPlace, True,
									 XtNstring, code_str, XtNlength, CODELEN,
									 NULL);
	cpsample = XtVaCreateManagedWidget("cpsample", commandWidgetClass, cpform,
									   NULL);
	
	cpokay = XtVaCreateManagedWidget("cpokay", commandWidgetClass, cpform,
									 NULL);
	cpcinput = XtVaCreateManagedWidget("cpcinput", commandWidgetClass, cpform,
									   NULL);
	cpcancel = XtVaCreateManagedWidget("cpcancel", commandWidgetClass, cpform,
									   NULL);
	cpminput = XtVaCreateManagedWidget("cpminput", menuButtonWidgetClass,
									   cpform, NULL);
	cplistshell = XtVaCreatePopupShell("cplistshell",
									   overrideShellWidgetClass, cpshell,
									   NULL);
	cplist = XtVaCreateManagedWidget("cplist", listWidgetClass, cplistshell,
									 NULL);

	menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass,
								cpminput, NULL);
	for (i = 1; i < 100 && i <= app_data.number_of_menu_items; i++) {
		Widget item;
		char item_name[10];

		sprintf(item_name, "item%d", i);
		item = XtVaCreateManagedWidget(item_name, smeBSBObjectClass,
									   menu, NULL);
		XtAddCallback(item, XtNcallback, (XtCallbackProc)ItemChosen, item);
	}
	strshell = XtVaCreatePopupShell("strshell", transientShellWidgetClass,
									menu, NULL);
	strlist = XtVaCreateManagedWidget("strlist", listWidgetClass, strshell,
									  NULL);

	XtAddCallback(cpkdf, XtNcallback, (XtCallbackProc)CPPopupFileList, KDFEXT);
	XtAddCallback(cpksm, XtNcallback, (XtCallbackProc)CPPopupFileList, KSMEXT);
	XtAddCallback(cplist, XtNcallback, (XtCallbackProc)CPFileChosen, NULL);
	XtAddCallback(cpsample, XtNcallback, CPCodeInput, NULL);
	XtAddCallback(cpokay, XtNcallback, CPOkay, NULL);
	XtAddCallback(cpcancel, XtNcallback, CPCancel, NULL);
	XtAddCallback(strlist, XtNcallback, (XtCallbackProc)StringChosen, NULL);
	
	XtVaGetValues(korEngStat,
				  XtNjohabFont, &jf, XtNinternalWidth, &iw,
				  XtNinternalHeight, &ih, NULL);
	XtVaSetValues(korEngStat,
				  XtNwidth, jf->max_bounds.width + 2*iw,
				  XtNheight,
				  jf->max_bounds.ascent+jf->max_bounds.descent + 2 * ih, NULL);
}

static void parse_resources()
{
	KeySym keysym;
	ModifierMaskType modifier, modifier_mask;
	
	if (app_data.transport == NULL &&
		app_data.toggle_key == NULL &&
		app_data.chinese_key == NULL &&
		app_data.popup_key == NULL &&
		app_data.flush_key == NULL)
		print_e(0, "The app-defaults file seems to be missing.\n");

	if (app_data.transport == NULL)
		print_e(1, "No transport mechanism specified.\n");
	if (app_data.toggle_key == NULL)
		print_e(1, "No Korean-English toggle key specified.\n");
	if (app_data.chinese_key == NULL)
		print_e(1, "No Chinese conversion key specified.\n");
	if (app_data.popup_key == NULL)
		print_e(1, "No Popup key specified.");
	if (app_data.flush_key == NULL)
		print_e(1, "No Flush Key specified.");
	
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.toggle_key))
		print_e(1, XtNkorEngToggleKey " resource malformed.\n");
	else {
		trigger_keys[TOGGLE].keysym = keysym;
		trigger_keys[TOGGLE].modifier = modifier;
		trigger_keys[TOGGLE].modifier_mask = modifier_mask;
	}
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.popup_key))
		print_e(1, XtNpopupKey " resource malformed.\n");
	else {
		trigger_keys[POPUP].keysym = keysym;
		trigger_keys[POPUP].modifier = modifier;
		trigger_keys[POPUP].modifier_mask = modifier_mask;
	}
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.chinese_key))
		print_e(1, XtNchineseConversionKey " resource malformed.\n");
	else {
		trigger_keys[CONVERSION].keysym = keysym;
		trigger_keys[CONVERSION].modifier = modifier;
		trigger_keys[CONVERSION].modifier_mask = modifier_mask;
	}
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.flush_key))
		print_e(1, XtNflushKey " resource malformed.\n");
	else {
		flush_key.keysym = keysym;
		flush_key.modifier = modifier;
		flush_key.modifier_mask = modifier_mask;
	}
}

static void open_om()
{
	XrmDatabase rdb = XtDatabase(display);
	char *name, *class;
	
	XtGetApplicationNameAndClass(display, &name, &class);
	if ((om = XOpenOM(display, rdb, name, class)) == NULL)
		print_e(1, "Cannot open output method.\n");
}
	
static void help_msg()
{
	fprintf(stderr, "%s understands all standard Xt command-line options.\n"
			"Additional options are as follows:\n"
			"-transport <transport connection name>\n"
			"-kdf <keyboard description file>\n"
			"-sortcd <chinese word dictionary file>\n",
			program_name);
}

static int protocol_handler(XIMS ims, IMProtocol *call_data)
{
	IContext *target;
	ICIDType icid;
	
	switch (call_data->major_code) {
	case XIM_OPEN:
		return open_handler(ims, &call_data->imopen);
	case XIM_CLOSE:
		return close_handler(ims, &call_data->imclose);
	case XIM_CREATE_IC:
		return create_ic_handler(ims, &call_data->changeic);
	case XIM_DESTROY_IC:
		return destroy_ic_handler(ims, &call_data->destroyic);
	default:
		// Each struct of the following protocols have icid as its
		// fourth field.  So icid will have the correct value whether
		// the request is XIM_RESET_IC or not.
		icid = call_data->resetic.icid;
		
		if ((target = [theICManager findICWithID:icid]))
			switch (call_data->major_code) {
			case XIM_SET_IC_FOCUS:
				return [target setICFocus:&call_data->changefocus];
			case XIM_UNSET_IC_FOCUS:
				return [target unsetICFocus:&call_data->changefocus];
			case XIM_RESET_IC:
				return [target resetIC:&call_data->resetic];
			case XIM_SET_IC_VALUES:
				return [target setICValues:&call_data->changeic];
			case XIM_GET_IC_VALUES:
				return [target getICValues:&call_data->changeic];
			case XIM_TRIGGER_NOTIFY:
				return [target triggerNotify:&call_data->triggernotify];
			case XIM_FORWARD_EVENT:
				return [target forwardEvent:&call_data->forwardevent];
			case XIM_PREEDIT_START_REPLY:
				return [target preeditStartReply:&call_data->preedit_callback];
			case XIM_PREEDIT_CARET_REPLY:
				return [target preeditCaretReply:&call_data->preedit_callback];
			case XIM_STR_CONVERSION_REPLY:
				return
					[target strConversionReply:&call_data->strconv_callback];
			default:
				print_e(0, "Unknown protocol number, %d, encountered.\n",
						call_data->major_code);
			}
		else
			print_e(0, "Strange IC id.\n");
	}
	
	return False;
}

static int open_handler(XIMS ims, IMOpenStruct *cd)
{
#ifdef DEBUG
	print_e(0, "open_handler(): connect_id = %d, lang = %s.\n",
		   (int)cd->connect_id, cd->lang.name);
#endif
	return True;
}

static int close_handler(XIMS ims, IMCloseStruct *cd)
{
#ifdef DEBUG
	print_e(0, "close_handler(): connect_id = %d.\n", (int)cd->connect_id);
#endif
	return True;
}

static int create_ic_handler(XIMS ims, IMChangeICStruct *cd)
{
	IContext *anIC = [[IContext alloc] initWith:cd];

	if (anIC == nil ||
		[theICManager add:anIC] == nil)
	{
		[anIC free];
		print_e(0, "Input context initialization failed.\n");
		return False;
	}

	// IC id is returned to the client through this structure.
	cd->icid = [anIC getID];
	
	return True;
}

static int destroy_ic_handler(XIMS ims, IMDestroyICStruct *cd)
{
	if ([lastFocusedIC getID] == cd->icid)
		lastFocusedIC = nil;
	[theICManager deleteICWithID:cd->icid];
#ifdef DEBUG
	print_e(0, "destroy_ic_handler(): IC of id %d destroyed.\n", cd->icid);
#endif
	return True;
}

//
//  utility functions
//

void app_main_loop(XtAppContext ac, BOOL *continue_looping)
{
	while (*continue_looping) {
		XEvent event;

		XtAppNextEvent(ac, &event);
		XtDispatchEvent(&event);
	}
}
	
int trigger_key_index(XKeyEvent *kev)
{
	int i;

	for (i = 0; i < XtNumber(trigger_keys); i++)
		if (is_trigger_key(kev, &trigger_keys[i]))
			return i;

	return -1;
}

BOOL is_trigger_key(XKeyEvent *kev, XIMTriggerKey *trigger)
{
	return XLookupKeysym(kev, 0) == trigger->keysym &&
		(kev->state & trigger->modifier_mask) == trigger->modifier;
}

void print_e(int exit_code, char *format, ...)
{
	va_list ap;

	if (format == NULL)
		perror(program_name);
	else {
		va_start(ap, format);
		fprintf(stderr, "%s: ", program_name);
		vfprintf(stderr, format, ap);
		va_end(ap);
	}

	if (exit_code != 0)
		exit(exit_code);
}

void CenterWidgetAtPoint(Widget w, Position x, Position y)
{
    Dimension   width, height;

	XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL);

    x = x - width / 2;
    y = y - height / 2;

    if (x < 0)
        x = 0;
    else {
        int scr_width = WidthOfScreen(XtScreen(w));

        if (x + width > scr_width)
            x = scr_width - width;
    }

    if (y < 0)
        y = 0;
    else {
        int scr_height = HeightOfScreen(XtScreen(w));

        if (y + height > scr_height)
            y = scr_height - height;
    }
	
	XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
}

void SetListWidth(Widget w)
{
	int longest;
	int number;
	Dimension column_spacing;
	Dimension internal_width;
	Dimension one_row_width;
	Dimension half_screen_width;
	
	XtVaGetValues(w, XtNlongest, &longest, XtNnumberStrings, &number,
				  XtNcolumnSpacing, &column_spacing,
				  XtNinternalWidth, &internal_width, NULL);
	one_row_width = (longest + column_spacing) * number + 2 * internal_width;
	half_screen_width = WidthOfScreen(XtScreen(w)) / 2;
 	XtVaSetValues(w, XtNwidth, MIN(one_row_width, half_screen_width), NULL);
}

//
//  callback procedures
//

/* ARGSUSED */
static void CPPopupFileList(Widget w,
							char *file_ext, XtPointer call_data)
{
	char *p, *q;
	char directory[FILENAME_MAX];
	
	FreeFileList();
	if ((file_list = malloc(3 * sizeof(file_list[0]))) == NULL)
		return;
	file_list_size = 3;
	
	for (nfiles = 0, p = app_data.search_path; *p; p = (*q ? q + 1 : q)) {
		DIR *ds;
		struct dirent *e;
		
		if ((q = strchr(p, ':')) == NULL)
			q = p + strlen(p);
		strmemcpy(directory, p, q - p);
		if ((ds = opendir(directory)) == NULL) {
			print_e(0, "Cannot open directory: %s\n", directory);
			continue;
		}
		while ((e = readdir(ds)) != NULL) {
			if (has_extension(e->d_name, file_ext) &&
				access(strcat(strcat(directory, "/"), e->d_name), R_OK) == 0)
			{
				if (nfiles >= file_list_size) {
					char **new_file_list;
					int new_file_list_size = file_list_size + file_list_size/2;

					if ((new_file_list =
						 realloc(file_list, new_file_list_size *
								 sizeof(new_file_list[0]))) == NULL)
						goto sort;
					file_list = new_file_list;
					file_list_size = new_file_list_size;
				}

				if ((file_list[nfiles] = strdup(e->d_name)) == NULL)
					goto sort;
				nfiles++;
			}
			directory[q - p] = '\0';
		}
		closedir(ds);
	} // for

sort:
	if (nfiles > 0) {
		Position x, y;
		Dimension width;
		static int compare_file_names(char **a, char **b);

		qsort(file_list, nfiles, sizeof(file_list[0]),
			  (QsortComp)compare_file_names);
		XawListChange(cplist, file_list, nfiles, 0, True);
		XtTranslateCoords(w, 0, 0, &x, &y);
		XtVaGetValues(w, XtNwidth, &width, NULL);
		XtVaSetValues(cplistshell, XtNx, x + width, XtNy, y, NULL);
		XtPopupSpringLoaded(cplistshell);
		XtGrabPointer(cplistshell, True, ButtonReleaseMask,
					  GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
	}
}

static int compare_file_names(char **a, char **b)
{
	return strcmp(*a, *b);
}

/* ARGSUSED */
static void CPFileChosen(Widget w, XtPointer client_data,
						 XawListReturnStruct *call_data)
{
	if (has_extension(call_data->string, KDFEXT))
		XtVaSetValues(cpkdf, XtNlabel, call_data->string, NULL);
	else if (has_extension(call_data->string, KSMEXT))
		XtVaSetValues(cpksm, XtNlabel, call_data->string, NULL);
}

/* ARGSUSED */
static void CPOkay(Widget w, XtPointer client_data, XtPointer call_data)
{
	[lastFocusedIC changeParameters];
	XtPopdown(cpshell);
	cp_popped_up = NO;
	FreeFileList();
}

/* ARGSUSED */
static void CPCancel(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(cpshell);
	cp_popped_up = NO;
	FreeFileList();
}

static void FreeFileList()
{
	if (file_list) {
		char **p;

		for (p = file_list; nfiles > 0; p++, nfiles--)
			free(*p);
		free(file_list);
		file_list = NULL;
		file_list_size = 0;
	}
}

/* ARGSUSED */
static void CPCodeInput(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *string;

	XtVaGetValues(cpsample, XtNlabel, &string, NULL);
	if (string[0])
		[lastFocusedIC insertString:string];
}

/* ARGSUSED */
static void ItemChosen(Widget w, Widget client_data, XtPointer call_data)
{
	XrmValue value;
	char *type;
	char resource_name[256];
	char resource_class[256];
	char *app_name, *app_class;
	
	XtGetApplicationNameAndClass(display, &app_name, &app_class);
	sprintf(resource_name, "%s.cpshell.cpform.cpminput.menu.%s.strings",
			app_name, XtName(client_data));
	sprintf(resource_class,
			"%s.TransientShell.Form.MenuButton.SimpleMenu.SmeBSB.Strings",
			app_class);
	if (XrmGetResource(XtDatabase(display), resource_name, resource_class,
					   &type, &value))
	{
		char *strs = value.addr;
		char *p, *q, *e;
		int n;
		char **s;
		Window root, child;
		int rx, ry, wx, wy;
		unsigned int mask;
		Window transient_for;
		
		for (n = 0, p = strs; *p; p++)
			if (*p == '\n')
				n++;
		e = p;					// end of strs
		if (p > strs && p[-1] != '\n')
			n++;

		FreeStringList();		// nstrings = 0
		if ((string_list = malloc((n+1) * sizeof(string_list[0]))) == NULL)
			return;

		for (s = string_list, p = strs; n > 0; s++, n--, p = q+1) {
			p = eat_space(p, e);
			q = next_occurrence_of('\n', p, e);
			if ((*s = malloc(q - p + 1)) == NULL)
				break;
			strmemcpy(*s, p, q - p);
			nstrings++;
		}
		*s = "";

		XawListChange(strlist, string_list, nstrings + 1, 0, True);
 		SetListWidth(strlist);

		XQueryPointer(display, XtWindow(topLevel),
					  &root, &child, &rx, &ry, &wx, &wy, &mask);
 		CenterWidgetAtPoint(strshell, rx, ry);
		XtPopup(strshell, XtGrabNone);
		XGetTransientForHint(display, XtWindow(cpshell), &transient_for);
 		XSetTransientForHint(display, XtWindow(strshell), transient_for);
	}
}

/* ARGSUSED */
static void StringChosen(Widget w, XtPointer client_data,
						 XawListReturnStruct *call_data)
{
	if (call_data->list_index == nstrings)
		XtPopdown(strshell);
	else
		if (call_data->string[0])
			[lastFocusedIC insertString:call_data->string];
}

static void FreeStringList()
{
	if (string_list) {
		char **p;

		for (p = string_list; nstrings > 0; p++, nstrings--)
			free(*p);
		free(string_list);
		string_list = NULL;
	}
}

//
//  action procedures
//

static void WMProtocols(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	if (ev->type == ClientMessage &&
        ev->xclient.message_type == wm_protocols &&
        ev->xclient.data.l[0] == wm_delete_window)
	{
        while (w && !XtIsShell(w))
            w = XtParent(w);
		
		if (w == topLevel)
			exit(0);			// FIXME!
        else if (w == cpshell)
			CPCancel(w, NULL, NULL);
        else if (w == [cd getShell])
			[cd popdown];
		else if (w == strshell)
			XtPopdown(strshell);
    }
}

static void ShowSample(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	unsigned long int code;
	char *endptr;
	char sample[3];
	
	code = strtoul(code_str, &endptr, 16);
	if (*endptr == '\0') {		// nice conversion
		if (code & 0xff00) {
			sample[0] = (code >> 8) & 0xff;
			sample[1] = code & 0xff;
			sample[2] = '\0';
		}
		else {
			sample[0] = code & 0xff;
			sample[1] = '\0';
		}
		XtVaSetValues(cpsample, XtNlabel, sample, NULL);
	}
	else
		XtVaSetValues(cpsample, XtNlabel, "", NULL);
}
